# -*- coding: utf-8 -*-

import logging
import os
from typing import Optional

from evernote_backup.cli_app_auth import get_auth_token, get_sync_client
from evernote_backup.cli_app_storage import get_storage, raise_on_old_database_version
from evernote_backup.cli_app_util import ProgramTerminatedError
from evernote_backup.config import CURRENT_DB_VERSION
from evernote_backup.note_exporter import NoteExporter, NothingToExportError
from evernote_backup.note_storage import initialize_db
from evernote_backup.note_synchronizer import NoteSynchronizer, WrongAuthUserError

logger = logging.getLogger(__name__)


def init_db(
    database: str,
    auth_user: Optional[str],
    auth_password: Optional[str],
    auth_is_oauth: bool,
    auth_token: Optional[str],
    force: bool,
    backend: str,
) -> None:
    if not auth_token:
        auth_token = get_auth_token(auth_user, auth_password, auth_is_oauth, backend)

    note_client = get_sync_client(auth_token, backend)

    logger.info("Initializing database {0}...".format(os.path.basename(database)))

    try:
        initialize_db(database, force)
    except FileExistsError:
        raise ProgramTerminatedError(
            "Database already exists."
            " Use --force option to overwrite existing database file."
        )

    storage = get_storage(database)

    new_user = note_client.user

    storage.config.set_config_value("DB_VERSION", str(CURRENT_DB_VERSION))
    storage.config.set_config_value("USN", "0")
    storage.config.set_config_value("auth_token", auth_token)
    storage.config.set_config_value("user", new_user)
    storage.config.set_config_value("backend", backend)

    logger.info(f"Successfully initialized database for {new_user}!")


def reauth(
    database: str,
    auth_user: Optional[str],
    auth_password: Optional[str],
    auth_is_oauth: bool,
    auth_token: Optional[str],
) -> None:
    storage = get_storage(database)

    raise_on_old_database_version(storage)

    backend = storage.config.get_config_value("backend")

    if not auth_token:
        auth_token = get_auth_token(auth_user, auth_password, auth_is_oauth, backend)

    note_client = get_sync_client(auth_token, backend)

    local_user = storage.config.get_config_value("user")

    if local_user != note_client.user:
        raise ProgramTerminatedError(
            f"Current user of this database is {local_user}, not {note_client.user}!"
            " Each user must use different database file."
        )

    storage.config.set_config_value("auth_token", auth_token)

    logger.info(f"Successfully refreshed auth token for {local_user}!")


def sync(database: str) -> None:
    storage = get_storage(database)

    raise_on_old_database_version(storage)

    backend = storage.config.get_config_value("backend")
    auth_token = storage.config.get_config_value("auth_token")

    note_client = get_sync_client(auth_token, backend)

    note_synchronizer = NoteSynchronizer(note_client, storage)

    try:
        note_synchronizer.sync()
    except WrongAuthUserError as e:
        raise ProgramTerminatedError(
            f"Current user of this database is {e.local_user}, not {e.remote_user}!"
            " Each user must use different database file."
        )

    logger.info("Synchronization completed!")


def export(
    database: str,
    single_notes: bool,
    include_trash: bool,
    output_path: str,
) -> None:
    storage = get_storage(database)

    raise_on_old_database_version(storage)

    exporter = NoteExporter(storage, output_path)

    try:
        exporter.export_notebooks(single_notes, include_trash)
    except NothingToExportError:
        raise ProgramTerminatedError(
            "Database is empty, nothing to export."
            " Execute sync command to populate database first!"
        )

    logger.info("All notes have been exported!")



# -*- coding: utf-8 -*-

import re
import uuid
from typing import Optional

import xmltodict
from evernote.edam.type.ttypes import Note, Resource

from evernote_backup.note_formatter_util import fmt_binary, fmt_content, fmt_time


class NoteFormatter(object):
    """https://xml.evernote.com/pub/evernote-export3.dtd"""

    def __init__(self) -> None:
        self._raw_elements: dict = {}

    def format_note(self, note: Note) -> str:
        self._raw_elements = {}

        note_skeleton = {
            "note": {
                "title": note.title,
                "created": fmt_time(note.created),
                "updated": fmt_time(note.updated),
                "tag": note.tagNames,
                "note-attributes": None,
                "content": self._fmt_raw(fmt_content(note.content)),
                "resource": map(self._fmt_resource, note.resources or []),
            }
        }

        if note.attributes:
            note_skeleton["note"]["note-attributes"] = {
                "subject-date": fmt_time(note.attributes.subjectDate),
                "latitude": note.attributes.latitude,
                "longitude": note.attributes.longitude,
                "altitude": note.attributes.altitude,
                "author": note.attributes.author,
                "source": note.attributes.source,
                "source-url": note.attributes.sourceURL,
                "source-application": note.attributes.sourceApplication,
                "reminder-order": note.attributes.reminderOrder,
                "reminder-time": note.attributes.reminderTime,
                "reminder-done-time": note.attributes.reminderDoneTime,
                "place-name": note.attributes.placeName,
                "content-class": note.attributes.contentClass,
            }

        note_template = xmltodict.unparse(
            note_skeleton,
            pretty=True,
            short_empty_elements=True,
            full_document=False,
            indent="  ",
            depth=1,
        )

        # Remove empty tags
        note_template = re.sub(r"^\s+<.*?/>\n", "", note_template, flags=re.M)

        for r_uuid, r_body in self._raw_elements.items():
            note_template = note_template.replace(r_uuid, r_body)

        return str(note_template)

    def _fmt_resource(self, resource: Resource) -> dict:
        return {
            "data": {
                "@encoding": "base64",
                "#text": self._fmt_raw(fmt_binary(resource.data.body)),
            },
            "mime": resource.mime,
            "width": resource.width,
            "height": resource.height,
            "duration": resource.duration,
            "resource-attributes": {
                "source-url": resource.attributes.sourceURL,
                "timestamp": fmt_time(resource.attributes.timestamp),
                "latitude": resource.attributes.latitude,
                "longitude": resource.attributes.longitude,
                "altitude": resource.attributes.altitude,
                "camera-make": resource.attributes.cameraMake,
                "camera-model": resource.attributes.cameraModel,
                "reco-type": resource.attributes.recoType,
                "file-name": resource.attributes.fileName,
                "attachment": resource.attributes.attachment,
            },
        }

    def _fmt_raw(self, body: Optional[str]) -> Optional[str]:
        if body is None:
            return body
        content_uuid = str(uuid.uuid4())
        self._raw_elements[content_uuid] = body
        return content_uuid



import threading
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Optional
from urllib.parse import parse_qsl, quote, urlparse

import oauth2

from evernote_backup.cli_app_util import is_inside_docker
from evernote_backup.config import OAUTH_LOCAL_PORT
from evernote_backup.evernote_client import EvernoteClientBase


class OAuthDeclinedError(Exception):
    """Raise when user cancels authentication"""


class CallbackHandler(BaseHTTPRequestHandler):
    http_codes = {
        "OK": 200,
        "NOT FOUND": 404,
    }

    def do_GET(self) -> None:
        response = urlparse(self.path)

        if response.path != "/oauth_callback":
            self.send_response(self.http_codes["NOT FOUND"])
            self.end_headers()
            return

        self.server.callback_response = dict(parse_qsl(response.query))  # type: ignore

        self.send_response(self.http_codes["OK"])
        self.end_headers()
        self.wfile.write(
            "<html><head><title>OAuth Callback</title></head>"
            "<body>You can close this tab now...</body></html>".encode("utf-8")
        )

    def log_message(self, *args, **kwargs) -> None:  # type: ignore
        """Silencing server log"""


class StoppableHTTPServer(HTTPServer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)

        self.callback_response: dict = {}

    def run(self) -> None:
        try:  # noqa: WPS501
            self.serve_forever()
        finally:
            self.server_close()


class EvernoteOAuthCallbackHandler(object):
    def __init__(self, oauth_client: "EvernoteOAuthClient") -> None:
        self.client = oauth_client

        self.server_host = "localhost"
        self.server_port = OAUTH_LOCAL_PORT
        self.oauth_token: dict = {}

    def get_oauth_url(self) -> str:
        self.oauth_token = self.client.get_request_token(
            f"http://{self.server_host}:{self.server_port}/oauth_callback"
        )

        return self.client.get_authorize_url(self.oauth_token)

    def wait_for_token(self) -> str:
        callback = self._wait_for_callback()

        if "oauth_verifier" not in callback:
            raise OAuthDeclinedError

        return self.client.get_access_token(
            oauth_token=callback["oauth_token"],
            oauth_verifier=callback["oauth_verifier"],
            oauth_token_secret=self.oauth_token["oauth_token_secret"],
        )

    def _wait_for_callback(self) -> dict:
        if is_inside_docker():
            server_param = ("0.0.0.0", self.server_port)  # noqa: S104
        else:
            server_param = (self.server_host, self.server_port)

        callback_server = StoppableHTTPServer(server_param, CallbackHandler)

        thread = threading.Thread(target=callback_server.run)
        thread.start()

        try:  # noqa: WPS501
            while not callback_server.callback_response:
                time.sleep(0.1)
        finally:
            callback_server.shutdown()
            thread.join()

        return callback_server.callback_response


class EvernoteOAuthClient(EvernoteClientBase):
    def __init__(
        self,
        backend: str,
        consumer_key: str,
        consumer_secret: str,
    ) -> None:
        super().__init__(backend=backend)

        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

    def get_authorize_url(self, request_token: dict) -> str:
        return "{0}?oauth_token={1}".format(
            self._get_endpoint("OAuth.action"),
            quote(request_token["oauth_token"]),
        )

    def get_request_token(self, callback_url: str) -> dict:

        client = self._get_oauth_client()

        request_url = "{0}?oauth_callback={1}".format(
            self._get_endpoint("oauth"), quote(callback_url)
        )

        _, response_content = client.request(request_url, "GET")

        return dict(parse_qsl(response_content.decode("utf-8")))

    def get_access_token(
        self,
        oauth_token: str,
        oauth_token_secret: str,
        oauth_verifier: str,
    ) -> str:
        token = oauth2.Token(oauth_token, oauth_token_secret)
        token.set_verifier(oauth_verifier)

        client = self._get_oauth_client(token)

        _, response_content = client.request(self._get_endpoint("oauth"), "POST")
        access_token_dict = dict(parse_qsl(response_content.decode("utf-8")))

        return access_token_dict["oauth_token"]

    def _get_oauth_client(self, token: Optional[oauth2.Token] = None) -> oauth2.Client:
        consumer = oauth2.Consumer(self.consumer_key, self.consumer_secret)

        return oauth2.Client(consumer, token) if token else oauth2.Client(consumer)



import base64
import time
from typing import Optional


def fmt_time(timestamp: Optional[int]) -> Optional[str]:
    if timestamp is None:
        return timestamp
    return time.strftime("%Y%m%dT%H%M%SZ", time.gmtime(timestamp // 1000))


def fmt_binary(binary_data: bytes) -> str:
    slice_width = 120
    return (
        "\n"
        + _slice_str(base64.b64encode(binary_data).decode(), slice_width)
        + "\n      "
    )


def fmt_content(content_body: Optional[str]) -> Optional[str]:
    if content_body is None:
        return content_body

    body = content_body.strip()

    # <?xml version="1.0" encoding="UTF-8"?>
    if body.startswith("<?xml") and body.find(">") != -1:
        content_start = body.find(">") + 1
        body = body[content_start:].strip()

    return (
        f"\n      "
        f"<![CDATA["
        f'<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
        f"{body}"
        f"]]>\n    "
    )


def _slice_str(text: str, width: int) -> str:
    slice_steps = range(0, len(text), width)
    slice_lines = (text[i : i + width] for i in slice_steps)
    return "\n".join(slice_lines)



# -*- coding: utf-8 -*-

import logging
import threading
from concurrent.futures import Future, ThreadPoolExecutor, as_completed, wait
from typing import Dict, Iterable, List, Tuple

from click import progressbar
from evernote.edam.type.ttypes import Note

from evernote_backup.cli_app_util import get_progress_output
from evernote_backup.config import SYNC_MAX_DOWNLOAD_WORKERS
from evernote_backup.evernote_client_sync import EvernoteClientSync
from evernote_backup.note_storage import SqliteStorage

logger = logging.getLogger(__name__)


class WrongAuthUserError(Exception):
    """Raise when remote auth user is not the same as the one registered in database"""

    def __init__(self, local_user: str, remote_user: str) -> None:
        self.local_user = local_user
        self.remote_user = remote_user


class WorkerStopException(Exception):
    """Raise when workers are stopped"""


class NoteClientWorker(object):
    def __init__(self, token: str, backend: str) -> None:
        self.stop = False
        self.token = token
        self.backend = backend

        self._thread_data = threading.local()

    def __call__(self, note_id: str) -> Note:
        if self.stop:
            raise WorkerStopException

        try:
            note_client = self._thread_data.note_client
        except AttributeError:
            note_client = EvernoteClientSync(token=self.token, backend=self.backend)
            self._thread_data.note_client = note_client

        return note_client.get_note(note_id)


class NoteSynchronizer(object):
    def __init__(
        self,
        note_client: EvernoteClientSync,
        note_storage: SqliteStorage,
    ) -> None:
        self._count_updated_notebooks = 0
        self._count_updated_notes = 0

        self._count_expunged_notebooks = 0
        self._count_expunged_notes = 0

        self.note_client = note_client
        self.storage = note_storage

    def sync(self) -> None:
        self._raise_on_wrong_user()

        remote_usn = self.note_client.get_remote_usn()
        current_usn = int(self.storage.config.get_config_value("USN"))

        if remote_usn == current_usn:
            logger.info("Noting to sync, current database is up to date!")
        else:
            logger.info("Syncing latest changes...")
            self._sync_chunks(current_usn, remote_usn)

        notes_to_sync = self.storage.notes.get_notes_for_sync()

        if notes_to_sync:
            logger.info("{0} notes to download...".format(len(notes_to_sync)))
            self._download_scheduled_notes(notes_to_sync)

            self._count_updated_notes = len(notes_to_sync)

        report = [
            f"Updated or added notebooks: {self._count_updated_notebooks}",
            f"Updated or added notes: {self._count_updated_notes}",
            f"Expunged notebooks: {self._count_expunged_notebooks}",
            f"Expunged notes: {self._count_expunged_notes}",
        ]

        for msg in report:
            logger.info(msg)

    def _raise_on_wrong_user(self) -> None:
        remote_user = self.note_client.user
        local_user = self.storage.config.get_config_value("user")

        if remote_user != local_user:
            raise WrongAuthUserError(local_user, remote_user)

    def _sync_chunks(self, current_usn: int, remote_usn: int) -> None:
        last_usn = current_usn
        with progressbar(
            length=remote_usn - current_usn,
            show_pos=True,
            file=get_progress_output(),  # type: ignore
        ) as chunks_bar:
            for chunk in self.note_client.iter_sync_chunks(current_usn):
                if current_usn > 0:
                    self._expunge(chunk.expungedNotebooks, chunk.expungedNotes)

                if chunk.notebooks:
                    self.storage.notebooks.add_notebooks(chunk.notebooks)

                    self._count_updated_notebooks += len(chunk.notebooks)

                if chunk.notes:
                    self.storage.notes.add_notes_for_sync(chunk.notes)

                self.storage.config.set_config_value("USN", str(chunk.chunkHighUSN))

                chunks_bar.update(chunk.chunkHighUSN - last_usn)
                last_usn = chunk.chunkHighUSN

    def _expunge(
        self, expunged_notebooks: List[str], expunged_notes: List[str]
    ) -> None:
        if expunged_notebooks:
            self.storage.notebooks.expunge_notebooks(expunged_notebooks)

            self._count_expunged_notebooks += len(expunged_notebooks)

        if expunged_notes:
            self.storage.notes.expunge_notes(expunged_notes)

            self._count_expunged_notes += len(expunged_notes)

    def _download_scheduled_notes(
        self, notes_to_sync: Iterable[Tuple[str, str]]
    ) -> None:
        with ThreadPoolExecutor(max_workers=SYNC_MAX_DOWNLOAD_WORKERS) as executor:
            note_worker = NoteClientWorker(
                self.note_client.token,
                self.note_client.backend,
            )

            note_futures = {
                executor.submit(note_worker, note_guid): note_title
                for note_guid, note_title in notes_to_sync
            }

            try:
                self._progress_sync_notes(note_futures)
            except KeyboardInterrupt:
                note_worker.stop = True

                logger.warning("Aborting, please wait...")
                wait(note_futures)

                raise

    def _progress_sync_notes(
        self,
        note_futures: Dict[Future, str],
    ) -> None:
        notes = as_completed(note_futures)

        with progressbar(
            notes,
            length=len(note_futures),
            item_show_func=lambda x: note_futures[x] if x else "",
            show_pos=True,
            file=get_progress_output(),  # type: ignore
        ) as notes_bar:
            for note_f in notes_bar:
                note = note_f.result()

                self.storage.notes.add_note(note)



import functools
import inspect
import platform
from typing import Any, Optional, Union

from evernote.edam.error.ttypes import (
    EDAMErrorCode,
    EDAMSystemException,
    EDAMUserException,
)
from evernote.edam.notestore import NoteStore
from evernote.edam.userstore import UserStore
from evernote.edam.userstore.ttypes import AuthenticationResult
from thrift.protocol import TBinaryProtocol
from thrift.transport import THttpClient

from evernote_backup.evernote_client_classes import ClientV2
from evernote_backup.evernote_client_util import EvernoteAuthError, network_retry


def raise_auth_error(exception: Union[EDAMSystemException, EDAMUserException]) -> None:
    messages = {
        EDAMErrorCode.BAD_DATA_FORMAT: {"authenticationToken": "Wrong token format!"},
        EDAMErrorCode.INVALID_AUTH: {
            "username": "Username not found!",
            "password": "Invalid password!",
            "oneTimeCode": "Invalid one-time code!",
            "authenticationToken": "Invalid authentication token!",
        },
        EDAMErrorCode.AUTH_EXPIRED: {
            "authenticationToken": "Authentication token expired or revoked!",
            "password": (
                "Password login disabled. Password reset required!\n"
                "Most probably, you log in to Evernote with Google or Apple account."
                " Use --oauth option to log in."
            ),
        },
    }

    if isinstance(exception, EDAMSystemException):
        error_param = exception.message
    else:
        error_param = exception.parameter

    try:
        error = messages[exception.errorCode][error_param]
    except KeyError:
        return

    raise EvernoteAuthError(error)


class EvernoteClientBase(object):
    def __init__(self, backend: str) -> None:
        self.backend = backend

        backends = {
            "evernote": "www.evernote.com",
            "evernote:sandbox": "sandbox.evernote.com",
            "china": "app.yinxiang.com",
            "china:sandbox": "sandbox.yinxiang.com",
        }

        self.service_host = backends[self.backend]

        if self.backend.startswith("china"):
            self.user_agent = "YXScript Windows/603139;"
        else:
            self.user_agent = "ENScript Windows/309091;"

        self.device_description = "evernote-backup"
        if platform.node():
            self.device_description += " [{0}]".format(platform.node())

    def _get_endpoint(self, path: str = "") -> str:
        return f"https://{self.service_host}/{path}"


class EvernoteClient(EvernoteClientBase):
    def __init__(self, backend: str, token: str) -> None:
        super().__init__(backend=backend)

        self.token = token
        self._user: Optional[str] = None

    def verify_token(self) -> None:
        try:
            self.user_store.getUser()
        except (EDAMUserException, EDAMSystemException) as e:
            raise_auth_error(e)
            raise

    @property
    def user_store(self) -> "Store":
        user_store_uri = self._get_endpoint("edam/user")
        return Store(
            client_class=ClientV2,
            store_url=user_store_uri,
            token=self.token,
            user_agent=self.user_agent,
        )

    @property
    def note_store(self) -> "Store":
        note_store_uri = self.user_store.getNoteStoreUrl()
        return Store(
            client_class=NoteStore.Client,
            store_url=note_store_uri,
            token=self.token,
            user_agent=self.user_agent,
        )

    @property
    def user(self) -> str:
        if self._user is None:
            self._user = self.user_store.getUser().username
        return self._user


class EvernoteClientAuth(EvernoteClient):
    def __init__(
        self,
        backend: str,
        consumer_key: str,
        consumer_secret: str,
    ):
        super().__init__(backend=backend, token="")  # noqa: S106

        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

    def login(self, username: str, password: str) -> AuthenticationResult:
        try:
            return self.user_store.authenticateLongSessionV2(
                username=username,
                password=password,
                ssoLoginToken=None,
                consumerKey=self.consumer_key,
                consumerSecret=self.consumer_secret,
                deviceIdentifier=None,
                deviceDescription=self.device_description,
                supportsTwoFactor=True,
                supportsBusinessOnlyAccounts=True,
            )
        except EDAMUserException as e:
            raise_auth_error(e)
            raise

    def two_factor_auth(self, auth_token: str, ota_code: str) -> AuthenticationResult:
        try:
            return self.user_store.completeTwoFactorAuthentication(
                authenticationToken=auth_token,
                oneTimeCode=ota_code,
                deviceIdentifier=None,
                deviceDescription=self.device_description,
            )
        except EDAMUserException as e:
            raise_auth_error(e)
            raise


class Store(object):
    def __init__(
        self,
        client_class: Union[UserStore.Client, NoteStore.Client],
        store_url: str,
        user_agent: str,
        token: Optional[str] = None,
    ):
        self.token = token
        self.user_agent = user_agent

        self._client = self._get_thrift_client(client_class, store_url)

    def __getattr__(self, name: str) -> Any:
        target_method = getattr(self._client, name)

        if not callable(target_method):
            return target_method

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            target_method_retry = network_retry(target_method)

            org_args = inspect.getfullargspec(target_method).args
            if len(org_args) == len(args) + 1:
                return target_method_retry(*args, **kwargs)
            elif self.token and "authenticationToken" in org_args:
                skip_args = ["self", "authenticationToken"]
                arg_names = [i for i in org_args if i not in skip_args]
                return functools.partial(
                    target_method_retry,
                    authenticationToken=self.token,
                )(
                    **dict(list(zip(arg_names, args))),
                )

            return target_method_retry(*args, **kwargs)

        return wrapper

    def _get_thrift_client(
        self,
        client_class: Union[UserStore.Client, NoteStore.Client],
        url: str,
    ) -> Union[UserStore.Client, NoteStore.Client]:
        http_client = THttpClient.THttpClient(url)
        http_client.setCustomHeaders(
            {
                "User-Agent": self.user_agent,
                "x-feature-version": 3,
                "accept": "application/x-thrift",
                "cache-control": "no-cache",
            }
        )

        thrift_protocol = TBinaryProtocol.TBinaryProtocol(http_client)
        return client_class(thrift_protocol)



import time
from http.client import HTTPException
from typing import Any, Callable

from evernote_backup.config import NETWORK_ERROR_RETRY_COUNT


class EvernoteAuthError(Exception):
    """Evernote authentication error"""


def network_retry(fun: Callable) -> Callable:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        retry_count = NETWORK_ERROR_RETRY_COUNT

        for i in range(retry_count):
            try:
                return fun(*args, **kwargs)
            except (HTTPException, ConnectionError):
                if i == retry_count - 1:
                    raise
                time.sleep(0.5)

    return wrapper



from typing import Dict, Iterator, Optional

from evernote.edam.notestore import NoteStore
from evernote.edam.notestore.ttypes import SyncChunk
from evernote.edam.type.ttypes import Note

from evernote_backup.config import SYNC_CHUNK_MAX_RESULTS
from evernote_backup.evernote_client import EvernoteClient


class EvernoteClientSync(EvernoteClient):
    def __init__(self, backend: str, token: str) -> None:
        super().__init__(backend=backend, token=token)

        self._tags: Optional[dict] = None

    def get_note(self, note_guid: str) -> Note:
        note = self.note_store.getNote(
            note_guid, True, True, True, True  # noqa: WPS425
        )

        if note.tagGuids:
            note.tagNames = [self.tags[t] for t in note.tagGuids]

        return note

    def iter_sync_chunks(self, after_usn: int) -> Iterator[SyncChunk]:
        sync_filter = NoteStore.SyncChunkFilter(
            includeNotes=True,
            includeNoteResources=True,
            includeNoteAttributes=True,
            includeNotebooks=True,
            includeExpunged=True,
        )

        max_results = SYNC_CHUNK_MAX_RESULTS

        while True:
            chunk = self.note_store.getFilteredSyncChunk(
                after_usn, max_results, sync_filter
            )

            yield chunk

            after_usn = chunk.chunkHighUSN

            if chunk.chunkHighUSN == chunk.updateCount:
                return

    @property
    def tags(self) -> Dict[str, str]:
        if self._tags is None:
            self._tags = {t.guid: t.name for t in self.note_store.listTags()}
        return self._tags

    def get_remote_usn(self) -> int:
        return int(self.note_store.getSyncState().updateCount)



# -*- coding: utf-8 -*-

import lzma
import os
import pickle
import sqlite3
from typing import Iterable, Iterator, Tuple, Union

from evernote.edam.type.ttypes import Note, Notebook

from evernote_backup.config import CURRENT_DB_VERSION

DB_SCHEMA = """CREATE TABLE IF NOT EXISTS notebooks(
                        guid TEXT PRIMARY KEY,
                        name TEXT,
                        stack TEXT
                    );
                    CREATE TABLE IF NOT EXISTS notes(
                        guid TEXT PRIMARY KEY,
                        title TEXT,
                        notebook_guid TEXT,
                        is_active BOOLEAN,
                        raw_note BLOB
                    );
                    CREATE TABLE IF NOT EXISTS config(
                        name TEXT PRIMARY KEY,
                        value TEXT
                    );
                    CREATE INDEX IF NOT EXISTS idx_notes
                     ON notes(notebook_guid, is_active);
                    CREATE INDEX IF NOT EXISTS idx_notes_title
                     ON notes(title COLLATE NOCASE);"""


class DatabaseResyncRequiredError(Exception):
    """Raise when database update requires resync"""


def initialize_db(filename: str, force: bool = False) -> None:
    if os.path.exists(filename):
        if force:
            if os.path.exists(filename) and force:
                os.remove(filename)
        else:
            raise FileExistsError

    db = sqlite3.connect(filename)

    with db as con:
        con.executescript(DB_SCHEMA)

    db.close()


class SqliteStorage(object):
    def __init__(self, database: Union[str, sqlite3.Connection]) -> None:
        if isinstance(database, sqlite3.Connection):
            self.db = database
        else:
            if not os.path.exists(database):
                raise FileNotFoundError("Database file does not exist.")

            self.db = sqlite3.connect(database)
            self.db.row_factory = sqlite3.Row

    @property
    def config(self) -> "ConfigStorage":
        return ConfigStorage(self.db)

    @property
    def notes(self) -> "NoteStorage":
        return NoteStorage(self.db)

    @property
    def notebooks(self) -> "NoteBookStorage":
        return NoteBookStorage(self.db)

    def check_version(self) -> None:
        try:
            db_version = int(self.config.get_config_value("DB_VERSION"))
        except KeyError:
            db_version = 0

        if db_version != CURRENT_DB_VERSION:
            self.upgrade_db(db_version)

    def upgrade_db(self, db_version: int) -> None:
        need_resync = False

        if db_version == 0:
            need_resync = True
            with self.db as con1:
                con1.execute("DROP TABLE notebooks;")
                con1.execute("DROP TABLE notes;")

                con1.executescript(DB_SCHEMA)

        if db_version < 3:
            with self.db as con2:
                con2.execute(
                    "CREATE INDEX IF NOT EXISTS idx_notes_title"
                    " ON notes(title COLLATE NOCASE);"
                )

        self.config.set_config_value("DB_VERSION", str(CURRENT_DB_VERSION))

        if need_resync:
            self.config.set_config_value("USN", "0")
            raise DatabaseResyncRequiredError


class NoteBookStorage(SqliteStorage):
    def add_notebooks(self, notebooks: Iterable[Notebook]) -> None:
        with self.db as con:
            con.executemany(
                "replace into notebooks(guid, name, stack) values (?, ?, ?)",
                ((nb.guid, nb.name, nb.stack) for nb in notebooks),
            )

    def iter_notebooks(self) -> Iterator[Notebook]:
        with self.db as con:
            cur = con.execute(
                "select guid, name, stack from notebooks",
            )

            yield from (
                Notebook(
                    guid=row["guid"],
                    name=row["name"],
                    stack=row["stack"],
                )
                for row in cur
            )

    def get_notebook_notes_count(self, notebook_guid: str) -> int:
        with self.db as con:
            cur = con.execute(
                "select COUNT(guid) from notes"
                " where notebook_guid=? and is_active=1",
                (notebook_guid,),
            )

            return int(cur.fetchone()[0])

    def expunge_notebooks(self, guids: Iterable[str]) -> None:
        with self.db as con:
            con.executemany("delete from notebooks where guid=?", ((g,) for g in guids))


class NoteStorage(SqliteStorage):
    def add_notes_for_sync(self, notes: Iterable[Note]) -> None:
        with self.db as con:
            con.executemany(
                "replace into notes(guid, title) values (?, ?)",
                ((n.guid, n.title) for n in notes),
            )

    def add_note(self, note: Note) -> None:
        note_deflated = lzma.compress(pickle.dumps(note))

        with self.db as con:
            con.execute(
                "replace into notes(guid, title, notebook_guid, is_active, raw_note)"
                " values (?, ?, ?, ?, ?)",
                (
                    note.guid,
                    note.title,
                    note.notebookGuid,
                    note.active,
                    note_deflated,
                ),
            )

    def iter_notes(self, notebook_guid: str) -> Iterator[Note]:
        with self.db as con:
            cur = con.execute(
                "select raw_note"
                " from notes where notebook_guid=? and is_active=1"
                " order by title COLLATE NOCASE",
                (notebook_guid,),
            )

            yield from (pickle.loads(lzma.decompress(row["raw_note"])) for row in cur)

    def iter_notes_trash(self) -> Iterator[Note]:
        with self.db as con:
            cur = con.execute(
                "select raw_note"
                " from notes where is_active=0"
                " order by title COLLATE NOCASE",
            )

            yield from (pickle.loads(lzma.decompress(row["raw_note"])) for row in cur)

    def get_notes_for_sync(self) -> Tuple[Tuple[str, str], ...]:
        with self.db as con:
            cur = con.execute("select guid, title from notes where raw_note is NULL")

            notes = ((row["guid"], row["title"]) for row in cur.fetchall())

            return tuple(notes)

    def expunge_notes(self, guids: Iterable[str]) -> None:
        with self.db as con:
            con.executemany("delete from notes where guid=?", ((g,) for g in guids))

    def get_notes_count(self, is_active: bool = True) -> int:
        with self.db as con:
            cur = con.execute(
                "select COUNT(guid) from notes where is_active=?", (is_active,)
            )

            return int(cur.fetchone()[0])


class ConfigStorage(SqliteStorage):
    def set_config_value(self, name: str, config_value: str) -> None:
        with self.db as con:
            con.execute(
                "replace into config(name, value) values (?, ?)",
                (name, config_value),
            )

    def get_config_value(self, name: str) -> str:
        with self.db as con:
            cur = con.execute("select value from config where name=?", (name,))
            res = cur.fetchone()

            if not res:
                raise KeyError(f"Config ID {name} not found in database!")

            return str(res[0])



API_DATA = b"Q0sJVEtPDWt7bmg2YnhtY344enZlMSRsPjxtaTI7P2Q0PTRhNWc="
API_DATA_OAUTH = b"eW91c3V3dn9mYjF2az48bzM7Pm4wZzdlZzpk"

NETWORK_ERROR_RETRY_COUNT = 50
OAUTH_LOCAL_PORT = 10500
SYNC_CHUNK_MAX_RESULTS = 200
SYNC_MAX_DOWNLOAD_WORKERS = 5
CURRENT_DB_VERSION = 3



import base64
import logging
import os
import sys
from typing import List, Optional

import click

logger = logging.getLogger(__name__)


class ProgramTerminatedError(Exception):
    """Terminate program with an error"""


def unscramble(scrambled_data: bytes) -> List[str]:
    scrambled_data = base64.b64decode(scrambled_data)

    unscrambled = b""
    for i, char in enumerate(scrambled_data):
        xor = len(scrambled_data) - i
        unscrambled += (char ^ xor).to_bytes(1, byteorder="big")

    return unscrambled.decode().split()


def get_progress_output() -> Optional[str]:
    if not is_console_interactive():
        return os.devnull

    return None


def is_console_interactive() -> bool:
    is_quiet = click.get_current_context().find_root().params["quiet"]

    return is_output_to_terminal() and not is_quiet


def is_output_to_terminal() -> bool:
    return sys.stdout.isatty()


def is_inside_docker() -> bool:
    return os.environ.get("INSIDE_DOCKER_CONTAINER", False) is not False



from datetime import datetime
from typing import Optional

from evernote_backup.cli_app_auth_oauth import evernote_login_oauth
from evernote_backup.cli_app_auth_password import evernote_login_password
from evernote_backup.cli_app_util import ProgramTerminatedError, logger
from evernote_backup.evernote_client_sync import EvernoteClientSync
from evernote_backup.evernote_client_util import EvernoteAuthError


def get_sync_client(auth_token: str, backend: str) -> EvernoteClientSync:
    logger.info(f"Authorizing auth token, {backend} backend...")

    client = EvernoteClientSync(token=auth_token, backend=backend)

    try:
        client.verify_token()
    except EvernoteAuthError as e:
        raise ProgramTerminatedError(e)

    token_expiration = get_token_expiration_date(auth_token)

    logger.info(f"Successfully authenticated as {client.user}!")
    logger.info(f"Current login will expire at {token_expiration}.")

    return client


def get_auth_token(
    auth_user: Optional[str],
    auth_password: Optional[str],
    auth_is_oauth: bool,
    backend: str,
) -> str:
    logger.info("Logging in to Evernote...")

    if auth_is_oauth:
        return evernote_login_oauth(backend)

    return evernote_login_password(auth_user, auth_password, backend)


def get_token_expiration_date(token: str) -> datetime:
    hex_base = 16
    token_parts = token.split(":")
    token_expire_ts = int(token_parts[2][2:], base=hex_base) // 1000

    return datetime.utcfromtimestamp(token_expire_ts)



from typing import Any, Callable, Iterable

import click

DIR_ONLY = click.Path(
    file_okay=False,
    writable=True,
    resolve_path=True,
)

FILE_ONLY = click.Path(
    dir_okay=False,
    writable=True,
    resolve_path=True,
)


class NaturalOrderGroup(click.Group):
    def list_commands(self, ctx: Any) -> Iterable:
        return self.commands.keys()


def group_options(*options: Callable) -> Callable:
    def wrapper(function: Callable) -> Callable:
        for option in reversed(options):
            function = option(function)
        return function

    return wrapper



import logging
import sys
import traceback
from typing import Optional

import click

from evernote_backup import cli_app
from evernote_backup.cli_app_click_util import (
    DIR_ONLY,
    FILE_ONLY,
    NaturalOrderGroup,
    group_options,
)
from evernote_backup.cli_app_util import ProgramTerminatedError, is_output_to_terminal
from evernote_backup.version import __version__

logger = logging.getLogger()


opt_user = click.option(
    "--user",
    "-u",
    help="Account email or username.",
)

opt_password = click.option(
    "--password",
    "-p",
    help="Account password.",
)

opt_oauth = click.option(
    "--oauth",
    is_flag=True,
    help=(
        "OAuth login flow."
        " Use this if you signed up for Evernote with Google or Apple account."
        " (Ignores '--user' and '--password' when used."
        " Doesn't work for Yinxiang.)"
    ),
)

opt_token = click.option(
    "--token",
    "-t",
    help=(
        "Manually provide authentication token to use with Evernote API."
        " (Advanced option, ignores '--user', '--password' and '--oauth' when used.)"
    ),
)

opt_database = click.option(
    "--database",
    "-d",
    default="en_backup.db",
    show_default=True,
    required=True,
    type=FILE_ONLY,
    help="Database file where notes are stored.",
)


@click.group(cls=NaturalOrderGroup)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="Quiet mode, output only in case of critical errors.",
)
@click.version_option(__version__)
def cli(quiet: bool) -> None:
    """Evernote backup & export

    \b
    On the first run, to export notes:
    evernote-backup init-db
    evernote-backup sync
    evernote-backup export output_dir/
    """

    if is_output_to_terminal():
        log_format = "%(message)s"
    else:
        log_format = "%(asctime)s | [%(levelname)s] | %(message)s"

    for h in logging.root.handlers[:]:
        logging.root.removeHandler(h)
        h.close()

    logging.basicConfig(format=log_format)

    if quiet:
        logger.setLevel(logging.CRITICAL)
    else:
        logger.setLevel(logging.INFO)


@cli.command()
@opt_database
@group_options(opt_user, opt_password, opt_oauth, opt_token)
@click.option(
    "--force",
    is_flag=True,
    help="Overwrite existing database file.",
)
@click.option(
    "--backend",
    default="evernote",
    show_default=True,
    type=click.Choice(["evernote", "evernote:sandbox", "china", "china:sandbox"]),
    help="API backend to connect to. If you are using Yinxiang, select 'china'.",
)
def init_db(
    database: str,
    user: Optional[str],
    password: Optional[str],
    oauth: bool,
    token: Optional[str],
    force: bool,
    backend: str,
) -> None:
    """Initialize storage & log in to Evernote."""

    cli_app.init_db(
        database=database,
        auth_user=user,
        auth_password=password,
        auth_is_oauth=oauth,
        auth_token=token,
        force=force,
        backend=backend,
    )


@cli.command()
@opt_database
def sync(database: str) -> None:
    """Sync local database with Evernote, downloading all notes."""

    cli_app.sync(database=database)


@cli.command()
@opt_database
@click.option(
    "--single-notes",
    is_flag=True,
    help="Export single notes instead of notebooks.",
)
@click.option(
    "--include-trash",
    is_flag=True,
    help="Include notes from trash into export.",
)
@click.argument(
    "output_path",
    required=True,
    type=DIR_ONLY,
)
def export(
    database: str, single_notes: bool, include_trash: bool, output_path: str
) -> None:
    """Export all notes from local database into ENEX files."""

    cli_app.export(
        database=database,
        single_notes=single_notes,
        include_trash=include_trash,
        output_path=output_path,
    )


click.password_option()


@cli.command()
@opt_database
@group_options(opt_user, opt_password, opt_oauth, opt_token)
def reauth(
    database: str,
    user: Optional[str],
    password: Optional[str],
    oauth: bool,
    token: Optional[str],
) -> None:
    """Refresh login to Evernote, run when token expires."""

    cli_app.reauth(
        database=database,
        auth_user=user,
        auth_password=password,
        auth_is_oauth=oauth,
        auth_token=token,
    )


def main() -> None:
    try:
        cli()
    except ProgramTerminatedError as e:
        logger.critical(e)
        sys.exit(1)
    except Exception:
        logger.critical(traceback.format_exc())
        sys.exit(1)



import click

from evernote_backup.cli_app_util import (
    ProgramTerminatedError,
    is_output_to_terminal,
    unscramble,
)
from evernote_backup.config import API_DATA_OAUTH
from evernote_backup.evernote_client_oauth import (
    EvernoteOAuthCallbackHandler,
    EvernoteOAuthClient,
    OAuthDeclinedError,
)


def get_oauth_client(backend: str) -> EvernoteOAuthClient:
    key, secret = unscramble(API_DATA_OAUTH)

    return EvernoteOAuthClient(
        consumer_key=key,
        consumer_secret=secret,
        backend=backend,
    )


def prompt_ota(delivery_hint: str) -> str:
    if not is_output_to_terminal():
        raise ProgramTerminatedError("Two-factor authentication requires user input!")

    one_time_hint = ""
    if delivery_hint:
        one_time_hint = " ({0})".format(delivery_hint)

    return str(click.prompt(f"Enter one-time code{one_time_hint}"))


def evernote_login_oauth(backend: str) -> str:
    if not is_output_to_terminal():
        raise ProgramTerminatedError("OAuth requires user input!")

    if backend.startswith("china"):
        raise ProgramTerminatedError(
            "OAuth is not supported to log in to Yinxiang!\n"
            "You will have to reset your Evernote account password"
            " and use your credentials to log in."
        )

    oauth_client = get_oauth_client(backend)

    oauth_handler = EvernoteOAuthCallbackHandler(oauth_client)

    oauth_url = oauth_handler.get_oauth_url()

    click.echo(
        f"Opening authorization page...\n"
        f"If it didn't open automatically, please copy this URL into your browser:\n"
        f"{oauth_url}"
    )
    click.launch(oauth_url)

    try:
        return oauth_handler.wait_for_token()
    except OAuthDeclinedError:
        raise ProgramTerminatedError("Authorization declined!")
